#!/home/brucef/code/PeateScience/local/dist/edosl0/0.1/env/bin/python
"""
Write a NetCDF file containing size in bytes of VIIRS scans.
"""
import os
from collections import defaultdict
from datetime import datetime

import netCDF4
import numpy as np
from edosl0util import jpssrdr
from grain import grain


band_map = {
    800: 'M5',
    801: 'M4',
    802: 'M3',
    803: 'M2',
    804: 'M1',
    805: 'M6',
    806: 'M7',
    807: 'M9',
    808: 'M10',
    809: 'M8',
    810: 'M11',
    811: 'M13',
    812: 'M12',
    813: 'I4',
    814: 'M16',
    815: 'M15',
    816: 'M14',
    817: 'I5',
    818: 'I1',
    819: 'I2',
    820: 'I3',
    821: 'DNB',
    825: 'CAL',
    826: 'ENGR'}
apids = sorted(band_map.keys())


def utc2tai(dt):
    return grain.Grain().utc2tai(dt, grain.VIIRS_EPOCH) * 1000**2


def create_netcdf_file(destpath):
    dataset = netCDF4.Dataset(destpath, 'w')
    dataset.bands = """Band APID  Day
==============
M1   804   X
M2   803   X
M3   802   X
M4   801   X
M5   800   X
M6   805   X
M7   806
M8   809
M9   807   X
M10  808
M11  810   X
M12  812
M13  811
M14  816
M15  815
M16  814

I1   818   X
I2   819   X
I3   820   X
I4   813
I5   817

DNB  821
CAL  825
ENGR 826"""
    dataset.createDimension('scan_time', None)
    dataset.createDimension('apid', size=len(apids))
    dataset.createVariable('time', 'u8', ('scan_time',))
    dataset.createVariable(
        'size', 'i4', ('apid', 'scan_time',),
        fill_value=-999, chunksizes=[len(apids), 1024])
    dataset.createVariable('apid', 'u2', ('apid',))

    dataset['apid'][:] = np.array(list(apids))

    return dataset


def read_data_from_rdr(filepath, start, end):
    # XXX: just get the VIIRS Science RDR for now
    rdr = jpssrdr.rdr_datasets(filepath)['science'][0]
    sizes = defaultdict(lambda: 0)
    times = set()
    for tracker, packet in rdr.packets():
        apid = packet.apid
        time = tracker.obs_time
        size = tracker.size
        if tracker.offset == -1:
            continue
        # skip data outside requested window
        if time < start or time >= end or apid not in apids:
            continue  # skip data outside window

        sizes[apid, time] += size
        times.add(time)

    return sorted(times), sizes


if __name__ == '__main__':
    import argparse, sys
    parser = argparse.ArgumentParser(description=__doc__)
    datetype = lambda v: datetime.strptime(v, '%Y-%m-%d %H:%M:%S')
    parser.add_argument('start', type=datetype)
    parser.add_argument('end', type=datetype)
    parser.add_argument('rdrs', type=argparse.FileType('r'), default=sys.stdin)
    args = parser.parse_args()

    start, end = utc2tai(args.start), utc2tai(args.end)
    destpath = args.start.strftime('viirs_scanbytes_d%Y%m%d_t%H%M%S.nc')
    dataset = create_netcdf_file(destpath)

    rdrfiles = sorted([l.strip() for l in args.rdrs],
                      key=os.path.basename)
    for filepath in rdrfiles:
        print "reading", filepath
        times, sizes = read_data_from_rdr(filepath, start, end)

        dat = np.ones((len(apids), len(times))) * -999
        for tidx, time in enumerate(times):
            for aidx, apid in enumerate(apids):
                dat[aidx][tidx] = sizes[apid, time]

        var = dataset['time']
        num_times = var.shape[0]
        var[num_times:] = np.array(list(times))

        var = dataset['size']
        var[:,num_times:] = dat

    dataset.close()
